/*
 *  IEC.cpp - IEC bus routines, 1541 emulation (DOS level)
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 *

 *
 * Notes:
 * ------
 *
 *  - There are three kinds of devices on the IEC bus: controllers,
 *    listeners and talkers. We are always the controller and we
 *    can additionally be either listener or talker. There can be
 *    only one listener and one talker active at the same time (the
 *    real IEC bus allows multiple listeners, but we don't).
 *  - There is one Drive object for every emulated drive (8..11).
 *    A pointer to one of them is stored in "listener"/"talker"
 *    when talk()/listen() is called and is used by the functions
 *    called afterwards.
 *  - The Drive objects have four virtual functions so that the
 *    interface to them is independent of their implementation:
 *      Open() opens a channel
 *      Close() closes a channel
 *      Read() reads from a channel
 *      Write() writes to a channel
 *  - The EOI/EOF signal is special on the IEC bus in that it is
 *    Sent before the last byte, not after it.
 */

#include "sysdeps.h"

#include "IEC.h"
#include "1541fs.h"
#include "1541d64.h"
#include "1541t64.h"
#include "Prefs.h"
#include "Display.h"
#include "C64.h"


/*
 *  Constructor: Initialize variables
 */

IEC::IEC(C64 *c64, C64Display *display) : the_c64(c64), the_display(display)
{
	// Create drive 8
	drive = NULL;	// Important because UpdateLEDs is called from the drive constructors (via set_error)

	if (!the_c64->ThePrefs.Emul1541Proc)
  {
		if (the_c64->ThePrefs.DriveType == DRVTYPE_DIR)
			drive = new FSDrive(the_c64, this, the_c64->armletData->disk);
		else if (the_c64->ThePrefs.DriveType == DRVTYPE_D64)
			drive = new D64Drive(the_c64, this, the_c64->armletData->disk);
		else
			drive = new T64Drive(the_c64, this, the_c64->armletData->disk);
		if (!drive)
			ShowRequester(REQ_NEW_DRIVE, 8, 0, 0, 0);
	}

	listener_active = talker_active = false;
	listening = false;
}


/*
 *  Destructor: Delete drives
 */

IEC::~IEC()
{
	if(NULL != drive)
		delete drive;
}


/*
 *  Reset all drives
 */

void IEC::Reset(void)
{
	if (drive != NULL && drive->Ready)
		drive->Reset();

	UpdateLEDs();
}


/*
 *  Preferences have changed, prefs points to new preferences,
 *  ThePrefs still holds the previous ones. Check if drive settings
 *  have changed.
 */

void IEC::NewPrefs(Prefs *prefs)
{
	// Delete and recreate all changed drives
	if ((the_c64->ThePrefs.DriveType != prefs->DriveType) 
	|| strcmp(the_c64->ThePrefs.DrivePath, prefs->DrivePath) 
	|| the_c64->ThePrefs.Emul1541Proc != prefs->Emul1541Proc)
	{
		if(drive != NULL)
		{
		  delete drive;
		  drive = NULL;	// Important because UpdateLEDs is called from drive constructors (via set_error())
		}
  
		if (!prefs->Emul1541Proc) {
			if (prefs->DriveType == DRVTYPE_DIR)
				drive = new FSDrive(the_c64, this, the_c64->armletData->disk);
			else if (prefs->DriveType == DRVTYPE_D64)
				drive = new D64Drive(the_c64, this, the_c64->armletData->disk);
			else
				drive = new T64Drive(the_c64, this, the_c64->armletData->disk);
		}
	}

	UpdateLEDs();
}


/*
 *  Update drive LED display
 */

inline void IEC::UpdateLEDs(void)
{
	if (drive != NULL)
		the_display->UpdateLEDs(drive->LED);
}


/*
 *  Output one byte
 */

uint8 IEC::Out(uint8 byte, bool eoi)
{
	if (listener_active) {
		if (received_cmd == CMD_OPEN)
			return open_out(byte, eoi);
		if (received_cmd == CMD_DATA)
			return data_out(byte, eoi);
		return ST_TIMEOUT;
	} else
		return ST_TIMEOUT;
}


/*
 *  Output one byte with ATN (Talk/Listen/Untalk/Unlisten)
 */

uint8 IEC::OutATN(uint8 byte)
{
	received_cmd = sec_addr = 0;	// Command is sent with secondary address
	switch (byte & 0xf0) {
		case ATN_LISTEN:
			listening = true;
			return listen(byte & 0x0f);
		case ATN_UNLISTEN:
			listening = false;
			return unlisten();
		case ATN_TALK:
			listening = false;
			return talk(byte & 0x0f);
		case ATN_UNTALK:
			listening = false;
			return untalk();
	}
	return ST_TIMEOUT;
}


/*
 *  Output secondary address
 */

uint8 IEC::OutSec(uint8 byte)
{
	if (listening) {
		if (listener_active) {
			sec_addr = byte & 0x0f;
			received_cmd = byte & 0xf0;
			return sec_listen();
		}
	} else {
		if (talker_active) {
			sec_addr = byte & 0x0f;
			received_cmd = byte & 0xf0;
			return sec_talk();
		}
	}
	return ST_TIMEOUT;
}


/*
 *  Read one byte
 */

uint8 IEC::In(uint8 *byte)
{
	if (talker_active && (received_cmd == CMD_DATA))
		return data_in(byte);

	*byte = 0;
	return ST_TIMEOUT;
}


/*
 *  Assert ATN (for Untalk)
 */

void IEC::SetATN(void)
{
	// Only needed for real IEC
}


/*
 *  Release ATN
 */

void IEC::RelATN(void)
{
	// Only needed for real IEC
}


/*
 *  Talk-attention turn-around
 */

void IEC::Turnaround(void)
{
	// Only needed for real IEC
}


/*
 *  System line release
 */

void IEC::Release(void)
{
	// Only needed for real IEC
}


/*
 *  Listen
 */

inline uint8 IEC::listen(int device)
{
	if (device == 8) {
		if ((listener = drive) != NULL && listener->Ready) {
			listener_active = true;
			return ST_OK;
		}
	}

	listener_active = false;
	return ST_NOTPRESENT;
}


/*
 *  Talk
 */

inline uint8 IEC::talk(int device)
{
	if (device == 8) {
		if ((talker = drive) != NULL && talker->Ready) {
			talker_active = true;
			return ST_OK;
		}
	}

	talker_active = false;
	return ST_NOTPRESENT;
}


/*
 *  Unlisten
 */

inline uint8 IEC::unlisten(void)
{
	listener_active = false;
	return ST_OK;
}


/*
 *  Untalk
 */

inline uint8 IEC::untalk(void)
{
	talker_active = false;
	return ST_OK;
}


/*
 *  Secondary address after Listen
 */

inline uint8 IEC::sec_listen(void)
{
	switch (received_cmd) {

		case CMD_OPEN:	// Prepare for receiving the file name
			name_ptr = name_buf;
			name_len = 0;
			return ST_OK;

		case CMD_CLOSE: // Close channel
			if (listener->LED != DRVLED_ERROR) {
				listener->LED = DRVLED_OFF;		// Turn off drive LED
				UpdateLEDs();
			}
			return listener->Close(sec_addr);
	}
	return ST_OK;
}


/*
 *  Secondary address after Talk
 */

inline uint8 IEC::sec_talk(void)
{
	return ST_OK;
}


/*
 *  Byte after Open command: Store character in file name, open file on EOI
 */

inline uint8 IEC::open_out(uint8 byte, bool eoi)
{
	if (name_len < NAMEBUF_LENGTH) {
		*name_ptr++ = byte;
		name_len++;
	}

	if (eoi) {
		*name_ptr = 0;				// End string
		listener->LED = DRVLED_ON;	// Turn on drive LED
		UpdateLEDs();
		return listener->Open(sec_addr, name_buf);
	}

	return ST_OK;
}


/*
 *  Write byte to channel
 */

inline uint8 IEC::data_out(uint8 byte, bool eoi)
{
	return listener->Write(sec_addr, byte, eoi);
}


/*
 *  Read byte from channel
 */

inline uint8 IEC::data_in(uint8 *byte)
{
	return talker->Read(sec_addr, byte);
}


/*
 *  Drive constructor
 */

Drive::Drive(C64 *c64, IEC *iec)
{
	the_c64 = c64;
	Errors_1541 = c64->armletData->errors1541;
	the_iec = iec;
	LED = DRVLED_OFF;
	Ready = false;
	set_error(ERR_STARTUP);
}


/*
 *  Set error message on drive
 */

void Drive::set_error(int error)
{
	error_ptr = Errors_1541[error];
	error_len = strlen(error_ptr);

	// Set drive condition
	if (error != ERR_OK)
		if (error == ERR_STARTUP)
			LED = DRVLED_OFF;
		else
			LED = DRVLED_ERROR;
	else if (LED == DRVLED_ERROR)
		LED = DRVLED_OFF;
	the_iec->UpdateLEDs();
}

// No virtual functions on palmos. Hilarity ensues.
uint8 Drive::Open(int channel, char *filename)
{
	switch (type) {
	    case TYPE_D64: return ((D64Drive *)this)->Open(channel, filename);
	    case TYPE_T64: return ((T64Drive *)this)->Open(channel, filename);
	    case TYPE_FS:  return ((FSDrive  *)this)->Open(channel, filename);
	}
	return 0;
}
uint8 Drive::Close(int channel)
{
	switch (type) {
	    case TYPE_D64: return ((D64Drive *)this)->Close(channel);
	    case TYPE_T64: return ((T64Drive *)this)->Close(channel);
	    case TYPE_FS:  return ((FSDrive  *)this)->Close(channel);
	}
	return 0;
}
uint8 Drive::Read(int channel, uint8 *byte)
{
	switch (type) {
	    case TYPE_D64: return ((D64Drive *)this)->Read(channel, byte);
	    case TYPE_T64: return ((T64Drive *)this)->Read(channel, byte);
	    case TYPE_FS:  return ((FSDrive  *)this)->Read(channel, byte);
	}
	return 0;
}
uint8 Drive::Write(int channel, uint8 byte, bool eoi)
{
	switch (type) {
	    case TYPE_D64: return ((D64Drive *)this)->Write(channel, byte, eoi);
	    case TYPE_T64: return ((T64Drive *)this)->Write(channel, byte, eoi);
	    case TYPE_FS:  return ((FSDrive  *)this)->Write(channel, byte, eoi);
	}
	return 0;
}
void Drive::Reset(void)
{
	switch (type) {
	    case TYPE_D64: return ((D64Drive *)this)->Reset();
	    case TYPE_T64: return ((T64Drive *)this)->Reset();
	    case TYPE_FS:  return ((FSDrive  *)this)->Reset();
	}
}
