/*
 *  SID.cpp - 6581 emulation
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 *

 *
 * Incompatibilities:
 * ------------------
 *
 *  - Lots of empirically determined constants in the filter calculations
 *  - Voice 3 cannot be muted
 *
 * PalmOS version:
 * All floating point math removed.
 * Lookup-tables externalized.
 * Global and static variables eliminated.
 */

#include "sysdeps.h"

#include "SID.h"
#include "Prefs.h"
#include "Display.h"
#include "VIC.h"


#define USE_FIXPOINT_MATHS
#define FIXPOINT_PREC 16	// number of fractional bits used in fixpoint representation
#define PRECOMPUTE_RESONANCE
#define ldSINTAB 9		// size of sinus table (0 to 90 degrees)

#include "FixPoint_PalmOS.i"


/*
 *  Constructor
 */

MOS6581::MOS6581(C64 *c64) : the_c64(c64)
{
	the_renderer = NULL;
	for (int i=0; i<32; i++)
		regs[i] = 0;

	// Open the renderer
	open_close_renderer(SIDTYPE_NONE, the_c64->ThePrefs.SIDType);
	//open_close_renderer(SIDTYPE_NONE, SIDTYPE_DIGITAL);
}


/*
 *  Destructor
 */

MOS6581::~MOS6581()
{
	// Close the renderer
	open_close_renderer(the_c64->ThePrefs.SIDType, SIDTYPE_NONE);
}


/*
 *  Reset the SID
 */

void MOS6581::Reset(void)
{
	for (int i=0; i<32; i++)
		regs[i] = 0;
	last_sid_byte = 0;

	// Reset the renderer
	if (the_renderer != NULL)
		the_renderer->Reset();
}


/*
 *  Preferences may have changed
 */

void MOS6581::NewPrefs(Prefs *prefs)
{
	open_close_renderer(the_c64->ThePrefs.SIDType, prefs->SIDType);
	if (the_renderer != NULL)
		the_renderer->NewPrefs(prefs);
}


/*
 *  Pause sound output
 */

void MOS6581::PauseSound(void)
{
	if (the_renderer != NULL)
		the_renderer->Pause();
}


/*
 *  Resume sound output
 */

void MOS6581::ResumeSound(void)
{
	if (the_renderer != NULL)
		the_renderer->Resume();
}


/*
 *  Get SID state
 */

void MOS6581::GetState(MOS6581State *ss)
{
	ss->freq_lo_1 = regs[0];
	ss->freq_hi_1 = regs[1];
	ss->pw_lo_1 = regs[2];
	ss->pw_hi_1 = regs[3];
	ss->ctrl_1 = regs[4];
	ss->AD_1 = regs[5];
	ss->SR_1 = regs[6];

	ss->freq_lo_2 = regs[7];
	ss->freq_hi_2 = regs[8];
	ss->pw_lo_2 = regs[9];
	ss->pw_hi_2 = regs[10];
	ss->ctrl_2 = regs[11];
	ss->AD_2 = regs[12];
	ss->SR_2 = regs[13];

	ss->freq_lo_3 = regs[14];
	ss->freq_hi_3 = regs[15];
	ss->pw_lo_3 = regs[16];
	ss->pw_hi_3 = regs[17];
	ss->ctrl_3 = regs[18];
	ss->AD_3 = regs[19];
	ss->SR_3 = regs[20];

	ss->fc_lo = regs[21];
	ss->fc_hi = regs[22];
	ss->res_filt = regs[23];
	ss->mode_vol = regs[24];

	ss->pot_x = 0xff;
	ss->pot_y = 0xff;
	ss->osc_3 = 0;
	ss->env_3 = 0;
}


/*
 *  Restore SID state
 */

void MOS6581::SetState(MOS6581State *ss)
{
	regs[0] = ss->freq_lo_1;
	regs[1] = ss->freq_hi_1;
	regs[2] = ss->pw_lo_1;
	regs[3] = ss->pw_hi_1;
	regs[4] = ss->ctrl_1;
	regs[5] = ss->AD_1;
	regs[6] = ss->SR_1;

	regs[7] = ss->freq_lo_2;
	regs[8] = ss->freq_hi_2;
	regs[9] = ss->pw_lo_2;
	regs[10] = ss->pw_hi_2;
	regs[11] = ss->ctrl_2;
	regs[12] = ss->AD_2;
	regs[13] = ss->SR_2;

	regs[14] = ss->freq_lo_3;
	regs[15] = ss->freq_hi_3;
	regs[16] = ss->pw_lo_3;
	regs[17] = ss->pw_hi_3;
	regs[18] = ss->ctrl_3;
	regs[19] = ss->AD_3;
	regs[20] = ss->SR_3;

	regs[21] = ss->fc_lo;
	regs[22] = ss->fc_hi;
	regs[23] = ss->res_filt;
	regs[24] = ss->mode_vol;

	// Stuff the new register values into the renderer
	if (the_renderer != NULL)
		for (int i=0; i<25; i++)
			the_renderer->WriteRegister(i, regs[i]);
}


/**
 **  Renderer for digital SID emulation (SIDTYPE_DIGITAL)
 **/

// TODO: stuff these inside DigitalRenderer
const uint32 SAMPLE_FREQ = 44100;	// Sample output frequency in Hz
const uint32 NYQUIST = 761;		// 1/(SAMPLE_FREQ/2) 8:24 fixpoint
const uint32 SID_FREQ = 985248;		// SID frequency in Hz
const uint32 CALC_FREQ = 50;		// Frequency at which calc_buffer is called in Hz (should be 50Hz)
const uint32 SID_CYCLES = SID_FREQ/SAMPLE_FREQ;	// # of SID clocks per sample frame
const int SAMPLE_BUF_SIZE = 0x138*2;	// Size of buffer for sampled voice (double buffered)

// SID waveforms (some of them :-)
enum {
	WAVE_NONE,
	WAVE_TRI,
	WAVE_SAW,
	WAVE_TRISAW,
	WAVE_RECT,
	WAVE_TRIRECT,
	WAVE_SAWRECT,
	WAVE_TRISAWRECT,
	WAVE_NOISE
};

// EG states
enum {
	EG_IDLE,
	EG_ATTACK,
	EG_DECAY,
	EG_RELEASE
};

// Filter types
enum {
	FILT_NONE,
	FILT_LP,
	FILT_BP,
	FILT_LPBP,
	FILT_HP,
	FILT_NOTCH,
	FILT_HPBP,
	FILT_ALL
};

// Structure for one voice
struct DRVoice {
	int wave;		// Selected waveform
	int eg_state;	// Current state of EG
	DRVoice *mod_by;	// Voice that modulates this one
	DRVoice *mod_to;	// Voice that is modulated by this one

	uint32 count;	// Counter for waveform generator, 8.16 fixed
	uint32 add;		// Added to counter in every frame

	uint16 freq;		// SID frequency value
	uint16 pw;		// SID pulse-width value

	uint32 a_add;	// EG parameters
	uint32 d_sub;
	uint32 s_level;
	uint32 r_sub;
	uint32 eg_level;	// Current EG level, 8.16 fixed

	uint32 noise;	// Last noise generator output value

	bool gate;		// EG gate bit
	bool ring;		// Ring modulation bit
	bool test;		// Test bit
	bool filter;	// Flag: Voice filtered

					// The following bit is set for the modulating
					// voice, not for the modulated one (as the SID bits)
	bool sync;		// Sync modulation bit
};

// Renderer class
class DigitalRenderer : public SIDRenderer {
public:
	DigitalRenderer(C64 *c64);
	~DigitalRenderer();

	void Reset(void);
	void EmulateLine(void);
	void WriteRegister(uint16 adr, uint8 byte);
	void NewPrefs(Prefs *prefs);
	void Pause(void);
	void Resume(void);

private:
	int32 *sintab;
	void init_sound(void);
	void calc_filter(void);
	void calc_buffer(int16 *buf, long count);

	uint8 Random();

	bool ready;				// Flag: Renderer has initialized and is ready
	uint8 volume;				// Master volume
	bool v3_mute;				// Voice 3 muted
	uint32 line;
	uint32 frame;

	const uint16 *TriTable;	// Tables for certain waveforms
	const uint16 *TriSawTable;
	const uint16 *TriRectTable;
	const uint16 *SawRectTable;
	const uint16 *TriSawRectTable;
	const uint32 *EGTable;	// Increment/decrement values for all A/D/R settings
	const uint8 *EGDRShift; // For exponential approximation of D/R
	const int16 *SampleTab; // Table for sampled voice
	const uint32 *resonanceLP;
	const uint32 *resonanceHP;

	DRVoice voice[3];				// Data for 3 voices

	uint8 f_type;					// Filter type
	uint8 f_freq;					// SID filter frequency (upper 8 bits)
	uint8 f_res;					// Filter resonance (0..15)
	uint32 seed;
	FixPoint f_ampl;
	FixPoint d1, d2, g1, g2;
	int32 xn1, xn2, yn1, yn2;		// can become very large
	FixPoint sidquot;

	uint8 sample_buf[SAMPLE_BUF_SIZE]; // Buffer for sampled voice
	uint32 sample_in_ptr;	// Index in sample_buf for writing

	C64 *the_c64;
};

/*
 *  Constructor
 */

DigitalRenderer::DigitalRenderer(C64 *c64) : the_c64(c64)
{
	// Link voices together
	voice[0].mod_by = &voice[2];
	voice[1].mod_by = &voice[0];
	voice[2].mod_by = &voice[1];
	voice[0].mod_to = &voice[1];
	voice[1].mod_to = &voice[2];
	voice[2].mod_to = &voice[0];

	TriTable=c64->armletData->triTable;
	TriSawTable=c64->armletData->triSawTable;
	TriRectTable=c64->armletData->triRectTable;
	SawRectTable=c64->armletData->sawRectTable;
	TriSawRectTable=c64->armletData->triSawRectTable;
	EGTable=c64->armletData->egTable;
	EGDRShift=c64->armletData->egdrShift;
	SampleTab=c64->armletData->sampleTab;
	sintab=c64->armletData->sintab;
	resonanceLP=c64->armletData->resonanceLP;
	resonanceHP=c64->armletData->resonanceHP;

	// Pre-compute the quotient. No problem since int-part is small enough
	// sidquot = (int32)((double)SID_FREQ)*65536) / SAMPLE_FREQ);
	// This isn't hardcoding, it's optimization through precomputation. Eh..
	sidquot = 1464154L;

	Reset();

	// System specific initialization
	init_sound();
}

uint8 DigitalRenderer::Random(void)
{
	seed = seed * 1103515245L + 12345;
	return seed >> 16;
}

/*
 *  Reset emulation
 */

void DigitalRenderer::Reset(void)
{
	volume = 0;
	v3_mute = false;
	seed = 1;

	for (int v=0; v<3; v++) {
		voice[v].wave = WAVE_NONE;
		voice[v].eg_state = EG_IDLE;
		voice[v].count = voice[v].add = 0;
		voice[v].freq = voice[v].pw = 0;
		voice[v].eg_level = voice[v].s_level = 0;
		voice[v].a_add = voice[v].d_sub = voice[v].r_sub = EGTable[0];
		voice[v].gate = voice[v].ring = voice[v].test = false;
		voice[v].filter = voice[v].sync = false;
	}

	f_type = FILT_NONE;
	f_freq = f_res = 0;
	f_ampl = 65536;
	d1 = d2 = g1 = g2 = 0;
	xn1 = xn2 = yn1 = yn2 = 0;

	sample_in_ptr = 0;
	memset(sample_buf, 0, SAMPLE_BUF_SIZE);
}


/*
 *  Write to register
 */

void DigitalRenderer::WriteRegister(uint16 adr, uint8 byte)
{
	if (!ready)
		return;

	int v = adr/7;	// Voice number

	switch (adr) {
		case 0:
		case 7:
		case 14:
			voice[v].freq = (voice[v].freq & 0xff00) | byte;
			voice[v].add = sidquot.imul((int)voice[v].freq);
			break;

		case 1:
		case 8:
		case 15:
			voice[v].freq = (voice[v].freq & 0xff) | (byte << 8);
			voice[v].add = sidquot.imul((int)voice[v].freq);
			break;

		case 2:
		case 9:
		case 16:
			voice[v].pw = (voice[v].pw & 0x0f00) | byte;
			break;

		case 3:
		case 10:
		case 17:
			voice[v].pw = (voice[v].pw & 0xff) | ((byte & 0xf) << 8);
			break;

		case 4:
		case 11:
		case 18:
			voice[v].wave = (byte >> 4) & 0xf;
			if ((byte & 1) != voice[v].gate)
				if (byte & 1)	// Gate turned on
					voice[v].eg_state = EG_ATTACK;
				else			// Gate turned off
					if (voice[v].eg_state != EG_IDLE)
						voice[v].eg_state = EG_RELEASE;
			voice[v].gate = byte & 1;
			voice[v].mod_by->sync = byte & 2;
			voice[v].ring = byte & 4;
			if ((voice[v].test = byte & 8))
				voice[v].count = 0;
			break;

		case 5:
		case 12:
		case 19:
			voice[v].a_add = EGTable[byte >> 4];
			voice[v].d_sub = EGTable[byte & 0xf];
			break;

		case 6:
		case 13:
		case 20:
			voice[v].s_level = (byte >> 4) * 0x111111;
			voice[v].r_sub = EGTable[byte & 0xf];
			break;

		case 22:
			if (byte != f_freq) {
				f_freq = byte;
				if (the_c64->ThePrefs.SIDFilters)
					calc_filter();
			}
			break;

		case 23:
			voice[0].filter = byte & 1;
			voice[1].filter = byte & 2;
			voice[2].filter = byte & 4;
			if ((byte >> 4) != f_res) {
				f_res = byte >> 4;
				if (the_c64->ThePrefs.SIDFilters)
					calc_filter();
			}
			break;

		case 24:
			volume = byte & 0xf;
			v3_mute = byte & 0x80;
			if (((byte >> 4) & 7) != f_type) {
				f_type = (byte >> 4) & 7;
				xn1 = xn2 = yn1 = yn2 = 0;
				if (the_c64->ThePrefs.SIDFilters)
					calc_filter();
			}
			break;
	}
}


/*
 *  Preferences may have changed
 */

void DigitalRenderer::NewPrefs(Prefs *prefs)
{
	calc_filter();
}


/*
 *  Calculate IIR filter coefficients
 */

void DigitalRenderer::calc_filter(void)
{
	FixPoint fr, arg;

	if (f_type == FILT_ALL)
	{
		d1 = 0; d2 = 0; g1 = 0; g2 = 0; f_ampl = 65536; return;
	}
	else if (f_type == FILT_NONE)
	{
		d1 = 0; d2 = 0; g1 = 0; g2 = 0; f_ampl = 0; return;
        }

	// Calculate resonance frequency
	if (f_type == FILT_LP || f_type == FILT_LPBP)
		fr = resonanceLP[f_freq];
	else
		fr = resonanceHP[f_freq];

	// Limit to <1/2 sample frequency, avoid div by 0 in case FILT_BP below
	// arg = fr / ((int32)SAMPLE_FREQ >> 1);
	// Try to avoid overflow. NYQUIST is 761, fr can be > 20000 (<<16)
	//arg = (fr.round() * NYQUIST)>>8;
	arg = (fr.Value() >> 8) * NYQUIST;

	if (arg > 64881) {arg = 64881;} // 0.99
	if (arg < 655) {arg = 655;}    // 0.01

	// Calculate poles (resonance frequency and resonance)
	//   0.55     1.2                   1   0.0133333333
	g2 = 36045 + 78643 * arg * (arg - 65536) + 874 * f_res;
	//      -2
	g1 = -131072 * g2.sqrt() * fixcos(arg, sintab);

	// Increase resonance if LP/HP combined with BP         0.1
	if (f_type == FILT_LPBP || f_type == FILT_HPBP) {g2 += 6554;}

	// Stabilize filter
	if (g1.abs() >= g2 + 1)
	{
	  if (g1 > 0) {g1 = g2 + 64881;} // 0.99
	  else {g1 = -(g2 + 64881);}
	}

	// Calculate roots (filter characteristic) and input attenuation
	switch (f_type)
	{
	  case FILT_LPBP:
	  case FILT_LP: // 2         1            0.25       1
		d1 = 131072; d2 = 65536; f_ampl = 16384 * (65536 + g1 + g2); break;
	  case FILT_HPBP:
	  case FILT_HP: // -2        1              0.25      1
		d1 = -131072; d2 = 65536; f_ampl = 16384 * (65536 - g1 + g2); break;
	  case FILT_BP:
		d1 = 0; d2 = -65536; // -1
		f_ampl = 16384 * (65536 + g1 + g2) * (65536 + fixcos(arg, sintab)) / fixsin(arg,sintab);
		break;
	  case FILT_NOTCH:
		d1 = -131072 * fixcos(arg, sintab); d2 = 65536;
		f_ampl = 16384 * (65536 + g1 + g2) * (65536 + fixcos(arg,sintab)) / fixsin(arg,sintab);
		break;
	  default: break;
	}
}


/*
 *  Fill one audio buffer with calculated SID sound
 */

void DigitalRenderer::calc_buffer(int16 *buf, long count)
{
	// Get filter coefficients, so the emulator won't change
	// them in the middle of our calculations
	FixPoint cf_ampl = f_ampl;
	FixPoint cd1 = d1, cd2 = d2, cg1 = g1, cg2 = g2;

	// Index in sample_buf for reading, 16.16 fixed
	uint32 sample_count = (sample_in_ptr + SAMPLE_BUF_SIZE/2) << 16;

	count >>= 1;	// 16 bit mono output, count is in bytes

	while (count--) {
		int32 sum_output;
		int32 sum_output_filter = 0;

		// Get current master volume from sample buffer,
		// calculate sampled voice
		uint8 master_volume = sample_buf[(sample_count >> 16) % SAMPLE_BUF_SIZE];
		sample_count += ((0x138 * 50L) << 16) / SAMPLE_FREQ;
		sum_output = SampleTab[master_volume] << 8;

		// Loop for all three voices
		for (int j=0; j<3; j++) {
			DRVoice *v = &voice[j];

			// Envelope generators
			uint16 envelope;

			switch (v->eg_state) {
				case EG_ATTACK:
					v->eg_level += v->a_add;
					if (v->eg_level > 0xffffff) {
						v->eg_level = 0xffffff;
						v->eg_state = EG_DECAY;
					}
					break;
				case EG_DECAY:
					if (v->eg_level <= v->s_level || v->eg_level > 0xffffff)
						v->eg_level = v->s_level;
					else {
						v->eg_level -= v->d_sub >> EGDRShift[v->eg_level >> 16];
						if (v->eg_level <= v->s_level || v->eg_level > 0xffffff)
							v->eg_level = v->s_level;
					}
					break;
				case EG_RELEASE:
					v->eg_level -= v->r_sub >> EGDRShift[v->eg_level >> 16];
					if (v->eg_level > 0xffffff) {
						v->eg_level = 0;
						v->eg_state = EG_IDLE;
					}
					break;
				case EG_IDLE:
					v->eg_level = 0;
					break;
			}
			envelope = (v->eg_level * master_volume) >> 20;

			// Waveform generator
			uint16 output;

			if (!v->test)
				v->count += v->add;

			if (v->sync && (v->count > 0x1000000))
				v->mod_to->count = 0;

			v->count &= 0xffffff;

			switch (v->wave) {
				case WAVE_TRI:
					if (v->ring)
						output = TriTable[(v->count ^ (v->mod_by->count & 0x800000)) >> 11];
					else
						output = TriTable[v->count >> 11];
					break;
				case WAVE_SAW:
					output = v->count >> 8;
					break;
				case WAVE_RECT:
					if (v->count > (uint32)(v->pw << 12))
						output = 0xffff;
					else
						output = 0;
					break;
				case WAVE_TRISAW:
					output = TriSawTable[v->count >> 16];
					break;
				case WAVE_TRIRECT:
					if (v->count > (uint32)(v->pw << 12))
						output = TriRectTable[v->count >> 16];
					else
						output = 0;
					break;
				case WAVE_SAWRECT:
					if (v->count > (uint32)(v->pw << 12))
						output = SawRectTable[v->count >> 16];
					else
						output = 0;
					break;
				case WAVE_TRISAWRECT:
					if (v->count > (uint32)(v->pw << 12))
						output = TriSawRectTable[v->count >> 16];
					else
						output = 0;
					break;
				case WAVE_NOISE:
					if (v->count > 0x100000) {
						output = v->noise = Random() << 8;
						v->count &= 0xfffff;
					} else
						output = v->noise;
					break;
				default:
					output = 0x8000;
					break;
			}
			if (v->filter)
				sum_output_filter += (int16)(output ^ 0x8000) * envelope;
			else
				sum_output += (int16)(output ^ 0x8000) * envelope;
		}

		// Filter
		if (the_c64->ThePrefs.SIDFilters) {
			int32 xn = cf_ampl.imul(sum_output_filter);
			int32 yn = xn+cd1.imul(xn1)+cd2.imul(xn2)-cg1.imul(yn1)-cg2.imul(yn2);
			yn2 = yn1; yn1 = yn; xn2 = xn1; xn1 = xn;
			sum_output_filter = yn;
		}

		*buf++ = (sum_output + sum_output_filter) >> 10;
	}
}


/*
 *  Open/close the renderer, according to old and new prefs
 */

void MOS6581::open_close_renderer(int old_type, int new_type)
{
	if (old_type == new_type)
		return;

	// Delete the old renderer
	delete the_renderer;

	// Create new renderer
	if (new_type == SIDTYPE_DIGITAL)
		the_renderer = new DigitalRenderer(the_c64);

	// Stuff the current register values into the new renderer
	if (the_renderer != NULL)
		for (int i=0; i<25; i++)
			the_renderer->WriteRegister(i, regs[i]);
}

// Lame switcher because of lack of virtual methods

void SIDRenderer::Reset(void)
{
	switch (type) {
		case SIDTYPE_DIGITAL: ((DigitalRenderer *)this)->Reset(); break;
	}
}
void SIDRenderer::EmulateLine(void)
{
	switch (type) {
		case SIDTYPE_DIGITAL: ((DigitalRenderer *)this)->EmulateLine(); break;
	}
}
void SIDRenderer::WriteRegister(uint16 adr, uint8 byte)
{
	switch (type) {
		case SIDTYPE_DIGITAL: ((DigitalRenderer *)this)->WriteRegister(adr, byte); break;
	}
}
void SIDRenderer::NewPrefs(Prefs *prefs)
{
	switch (type) {
		case SIDTYPE_DIGITAL: ((DigitalRenderer *)this)->NewPrefs(prefs); break;
	}
}
void SIDRenderer::Pause(void)
{
	switch (type) {
		case SIDTYPE_DIGITAL: ((DigitalRenderer *)this)->Pause(); break;
	}
}
void SIDRenderer::Resume(void)
{
	switch (type) {
		case SIDTYPE_DIGITAL: ((DigitalRenderer *)this)->Resume(); break;
	}
}


/*
 *  Initialization, create sub-process
 */
 
void DigitalRenderer::init_sound(void)
{
	ready=true;
	line=0;
	frame=0;
	type=SIDTYPE_DIGITAL;
}


/*
 *  Destructor, delete sub-process
 */

DigitalRenderer::~DigitalRenderer()
{
}


/*
 *  Sample volume (for sampled voice)
 */

void DigitalRenderer::EmulateLine(void)
{
	sample_buf[sample_in_ptr] = volume;
	sample_in_ptr = (sample_in_ptr + 1) % SAMPLE_BUF_SIZE;
	
	if (line == 0) {
		// TODO: fix hardcoded size. Size is in BYTES
		if(the_c64->armletData->soundBuf != NULL)
			calc_buffer(the_c64->armletData->soundBuf, 882*2);
		frame++;
		//the_c64->armletData->soundBuf[0]=frame;
	}
	line=(line+1)%TOTAL_RASTERS;
}


/*
 *  Pause sound output
 */

void DigitalRenderer::Pause(void)
{
}


/*
 *  Resume sound output
 */

void DigitalRenderer::Resume(void)
{
}
